"""An audio-friendly ls, with a little something extra"""
from datetime import datetime

__description__ = "an audio-friendly ls, with a little something extra"
__version__ = "0.1.7"
__version_repr__ = (
    f"sndls version {__version__} 2025-{datetime.now().year} developed by "
    "Esteban Gómez (Speech Interaction Technology, Aalto University)"
)
